/**************************************************
 *
  File: common-agent.cc
  Author: Suman Banerjee <suman@cs.umd.edu>
  Date: July 31, 2001
  Terms: GPL

  NICE implementation in myns

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 **************************************************/

#include <stdio.h>
#include <assert.h>
#include "nicenode.h"
#include "common-agent.h"
#include <scheduler.h>

//slee
#include "stdlib.h"

extern NamedConstant prob_data_loss;
NamedConstant const_cut_off("cut_off", 5000);
#define	PROB_DATA_LOSS  (prob_data_loss.getval().u.d)

//slee

#undef LOG_DETAILED_PKT_SEND
#undef LOG_DETAILED_PKT_RCV

void commonAgent::send_pkt_wrapper (AppPacket *ap, int aid, int nid) {

#ifdef LOG_DETAILED_PKT_SEND
   printf ("[ %d %d ] spd %f [ %d %d ] tp %d", id, n->id, Scheduler::Clock(), aid, nid, ap->st);
  display_detailed_pkt_info(ap);
#endif LOG_DETAILED_PKT_SEND
  if ( ap->st != PACKET_DATA ) {
	  m_sent_control++;
  }
  send_pkt(ap,aid,nid);
}

void commonAgent::rx_pkt_wrapper (AppPacket *ap) {
#ifdef LOG_DETAILED_PKT_RCV
   printf ("[ %d %d ] rpd %f [ %d %d ] tp %d", id, n->id, Scheduler::Clock(), ap->src_agent, ap->src, ap->st);
  display_detailed_pkt_info(ap);
#endif LOG_DETAILED_PKT_RCV
  return;
}

void commonAgent::display_detailed_pkt_info (AppPacket *ap) {

  switch (ap->st) {

  case JOIN_QUERY:
    printf (" sz 5 + 0\n");
    break;

  case JOIN_FORWARD:
    printf (" sz %d + 0\n", 2 * HOST_INFO_SIZE +  9);
    break;

  case PING_QUERY:
    printf (" sz 5 + 0\n");
    break;

  case JOIN_RESPONSE:
    printf (" sz %d + %d\n", HOST_INFO_SIZE + 7, HOST_INFO_SIZE * ap->u.joinresp_p.mbr_count);
    break;

  case CLUSTER_REFRESH:
    printf (" sz 13 + %d\n", HOST_INFO_SIZE * (ap->u.clusterrefresh_p.mbr_count+ap->u.clusterrefresh_p.hl_mbr_count));
    break;

  case CLUSTER_MERGE:
    printf (" sz 5 + %d\n", HOST_INFO_SIZE * ap->u.clustermerge_p.mbr_count);
    break;

  case PING_RESPONSE:
    printf (" sz 9 + %d\n", HOST_INFO_SIZE * ap->u.pingresp_p.mbr_count);
    break;

  case PACKET_DATA:
    printf(" sz %d + 0\n", 2 + HOST_INFO_SIZE);
    break;

  default:
    
    assert(0);
  }

  return;
}

//slee

int IdCache::getAgentId( int ni ) {
	if ( ni >= 0 && recent_nid == ni )
		return recent_aid;

	if ( ni >= 0 )
		for ( int i = 0; i < count; i++ ) {
			if ( nid[i] == ni ) {
				recent_aid = aid[i];
				recent_nid = nid[i];
				return aid[i];
			}
		}

	if ( count <= 0 )
		return -1;

	int index = -1;
	if ( USE_CORRELATION ) {
		if ( ni == MISS_BASED_CORR && sum_miss > 0 ) {
			int rnum = rand() % sum_miss;
			for ( int i = 0 ; i < count; i++ ) {
				rnum -= miss_score[i];
				if ( rnum < 0 ) {
					index = i;
					break;
				}
			}
		}
		if ( ni == RECV_BASED_CORR && sum_recv > 0 ) {
			int rnum = rand() % sum_recv;
			for ( int i  = 0 ; i  < count; i ++ ) {
				rnum -= recv_score[i];
				if ( rnum < 0 ) {
					index = i;
					break;
				}
			}
		}
	}

	if ( index < 0 ) {
	 	index = rand() % count;
	}
	recent_aid = aid[index];
	recent_nid = nid[index];
	return recent_aid;
}


int IdCache::getNodeId( int ai ) {
	if ( ai >= 0 && recent_aid == ai )
		return recent_nid;

	if ( ai >= 0 )
		for ( int i = 0; i < count; i++ ) {
			if ( aid[i] == ai ) {
				recent_aid = aid[i];
				recent_nid = nid[i];
				return nid[i];
			}
		}

	if ( count <= 0 )
		return -1;

	int index = rand() % count;
	recent_aid = aid[index];
	recent_nid = nid[index];
	return recent_nid;
}

int IdCache::addId( AppPacket *ap, int in_rscore, int in_mscore ) {
	query_random *ptr = &ap->u.random_p;
	return addId( ptr->rand_id.agent_id, ptr->rand_id.node_id, ptr->level, in_rscore, in_mscore );
}

int IdCache::addId(int ai, int ni, int lvl, int in_rscore, int in_mscore ) {

	int index = found(ai, ni);
	if ( index >= 0 ) {
		sum_recv -= recv_score[index];
		recv_score[index] = in_rscore;
		sum_recv += recv_score[index];

		sum_miss -= miss_score[index];
		miss_score[index] = in_mscore;
		sum_miss += miss_score[index];
	}
	//not found
	else if ( count < MAX_ID_CACHE ) {
		aid[count] = ai;
		nid[count] = ni;

		recv_score[count] = in_rscore;
		miss_score[count] = in_mscore;
		level[count] = 1 << (2 *lvl);

		sum_level += level[count];
		sum_recv += recv_score[count];
		sum_miss += miss_score[count];
		count++;
	}
	else {
		int index = getId();
		aid[index] = ai;
		nid[index] = ni;

		sum_level -= level[index];
		sum_recv -= recv_score[index];
		sum_miss -= miss_score[index];

		recv_score[count] = in_rscore;
		miss_score[count] = in_mscore;
		level[index] = 1 << (2 * lvl);

		sum_level += level[index];
		sum_recv += recv_score[index];
		sum_miss += miss_score[index];
	}

	return 1;
}

int IdCache::getId() {

	int i;
	for ( i = 0; i < count; i++ ) {
		if ( recv_score[i] == 0 && miss_score[i] == 0 )
			return i;
	}

	//fprintf(stderr, "non-zero id\n");
	return rand() % count;
}

int IdCache::found( int ai, int ni) {
	int i;
	for ( i = 0 ; i < count; i++ ) 
		if ( aid[i] ==  ai && nid[i] == ni )
			return i;

	return -1;
}

void IdCache::dump() {
	printf("----------------------\n");
	printf("sum_recv: %d, sum_miss: %d\n", sum_recv, sum_miss);
	for ( int i = 0 ; i < count; i++ ) {
		printf("\t%5d %5d %5d %5d\n", aid[i], nid[i], recv_score[i], miss_score[i]);
	}
}

////////////////////////////////////

void RxmitCache::clear() {
	memset( valid, 0, RXMIT_CACHE_SIZE ); 
}

int RxmitCache::lookup( int seq , double now) {
	int index = seq % RXMIT_CACHE_SIZE;
	if ( valid[index] && seq_no[index] == seq && timeout[index] > now )
		return 1;
	return 0;
}

int RxmitCache::insert( int seq, double rtt ) {
	assert ( rtt >= 0.0 );
	double now = Scheduler::Clock();

	if ( lookup( seq, now ) )
		return 1;

	int index = seq % RXMIT_CACHE_SIZE;
	valid[index] = 1;
	seq_no[index] = seq;
	timeout[index] = now + rtt * RXMIT_MULT;

	//fprintf(stderr, "%f %f %f\n", now, rtt, timeout[index] );

	return 0;
}
		
	
//slee
