/**************************************************
 *
  File: common-agent.h
  Author: Suman Banerjee <suman@cs.umd.edu>
  Date: July 31, 2001
  Terms: GPL

  NICE implementation in myns

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 **************************************************/

#ifndef _COMMON_AGENT_H_
#define _COMMON_AGENT_H_

#include <agent.h>
#include "app-packet.h"

//slee

extern NamedConstant prob_rand_send;
extern NamedConstant query_random_node_timeout;
extern NamedConstant const_max_bitmap_size;
extern NamedConstant const_proactive_increment;
extern NamedConstant const_use_correlation;

extern NamedConstant const_block_size;
extern NamedConstant const_rand_send_unit;
extern NamedConstant const_show_all;
extern NamedConstant const_cut_off;
extern NamedConstant const_use_my_random;

#define     PROB_RAND_SEND  (prob_rand_send.getval().u.d)
#define     QUERY_RANDOM_NODE_TIMEOUT   (query_random_node_timeout.getval().u.d)
#define     BITMAP_SIZE     (const_max_bitmap_size.getval().u.i)
#define     RXMIT_MULT     (const_proactive_increment.getval().u.i)
#define		USE_CORRELATION			(const_use_correlation.getval().u.i)

#define		BLOCK_SIZE		(const_block_size.getval().u.i)
#define		SHOW_ALL			(const_show_all.getval().u.i)

#define     USE_MY_RANDOM_SELECTION     (const_use_my_random.getval().u.i)
#define     DETERMINISTIC_PROACTIVE     1
#define     PROB_CONTINUE   (0.8)

#define     MAX_ITER        (10)
#define     FLAT_PROACTIVE  0

#define     FRESH_DATA      0 
#define     RAND_SEND_DATA  1
#define     PROACTIVE_DATA  2

#define     CUT_OFF     (const_cut_off.getval().u.i)
#define		LONG_TIME	(1000000.0)

#define     MAX_LATENCY             (15)

#define     RANDOM_SEND_LID     (-1)
#define     PROACTIVE_SEND_LID  (-2)
#define     ADOPT_LID		  	(-3)

#define		MISS_BASED_CORR		(-1)
#define		RECV_BASED_CORR		(-2)

//slee


class commonAgent : public Agent {

public :
	commonAgent() { m_sent_control = 0;}

  void send_pkt_wrapper (AppPacket *ap, int aid, int nid);
  void rx_pkt_wrapper (AppPacket *ap);

  void display_detailed_pkt_info (AppPacket *ap);

  int m_sent_control;

};


//slee

#define MAX_ID_CACHE	(100)
#define	RXMIT_CACHE_SIZE	(512)
#define	RAND_NUM_AT_A_TIME	(const_rand_send_unit.getval().u.i)

class IdCache {
	int recent_aid;
	int recent_nid;
	int count;

	int aid[MAX_ID_CACHE];
	int nid[MAX_ID_CACHE];
	int level[MAX_ID_CACHE];
	int recv_score[MAX_ID_CACHE];
	int miss_score[MAX_ID_CACHE];
	int sum_level;
	int sum_recv;
	int sum_miss;

	int found( int ai, int ni);
	int getId();

public:
	IdCache() { clear(); }
	//recent_nid = recent_aid = -1; count = sum_level = sum_score = 0; }

	int getAgentId(int i);
	int getNodeId(int i);

	int addId( int ai, int ni, int lvl, int in_rscore, int in_mscore);
	int addId( AppPacket *, int in_rscore, int in_mscore);

	void clear() { recent_nid = recent_aid = -1; count = sum_level = sum_recv = sum_miss = 0 ; }
	int getSize() { return count; }

	void dump();
};

class RxmitCache {
	char valid[RXMIT_CACHE_SIZE];
	int seq_no[RXMIT_CACHE_SIZE];
	double timeout[RXMIT_CACHE_SIZE];
public:
	RxmitCache() { clear(); }
	void clear();	// { memset ( valid, 0, RXMIT_CACHE_SIZE); }
	int lookup( int seq, double now );
	int insert( int seq, double RTT );
};


//slee

	
	
#endif _COMMON_AGENT_H_
  
