/**************************************************
 *
  File: nicenode.cc
  Author: Suman Banerjee <suman@cs.umd.edu>
  Date: July 31, 2001
  Terms: GPL

  NICE implementation in myns

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 **************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <constants.h>
#include "nicenode.h"
#include "app-packet.h"
#include "common.h"

//slee
#include "common-agent.h"
//slee

#define MAX_DATA_PACKETS (max_data_packets.getval().u.i)
#define	LINK_LOSS_PROB (const_loss_prob.getval().u.d)
#define	USE_TS_LOSS 	(const_use_ts_loss.getval().u.i)

NamedConstant max_data_packets("max_data_packets",5000);
NamedConstant const_loss_prob("loss_prob", 0.005);
NamedConstant const_use_ts_loss("use_ts_loss", 1);

//slee
int data_pkt_count;
int control_pkt_count;

static int initialized;


NiceNode::NiceNode (void) : Node () {

  m_data_packet_count = 0;
  m_control_packet_count = 0;
  m_control_byte_count = 0;

  if (MAX_DATA_PACKETS > 0)
    m_per_node_data_pkt_count = (int *) safe_malloc (sizeof(int) * MAX_DATA_PACKETS);
  else
    m_per_node_data_pkt_count = NULL;

  for (int i = 0; i < MAX_DATA_PACKETS; i++)
    m_per_node_data_pkt_count[i] = 0;

//slee
	m_loss_p = LINK_LOSS_PROB;
	m_use_ts_loss = USE_TS_LOSS;
//slee

}

NiceNode::~NiceNode (void) {

  if (m_per_node_data_pkt_count != NULL)
    free(m_per_node_data_pkt_count);
}

void NiceNode::specific_tx_pkt (Packet *p, int dst) {

	return;
  AppPacket *ap = (AppPacket *)p;
  if (ap->st == PACKET_DATA)
    printf ("[link-data-send] [ %d %d ] at %f seq %d\n", id, dst, Scheduler::Clock(), ap->u.data_p.seq_no);
  return;
}

void NiceNode::specific_rx_pkt_handler (Packet *p, bool is_src) {

  AppPacket *ap = (AppPacket *)p;

  switch (ap->st) {

  case PACKET_DATA:

    m_data_packet_count ++;
//slee
	data_pkt_count++;
//slee
    assert (ap->u.data_p.seq_no >= 0);
    if (ap->u.data_p.seq_no < MAX_DATA_PACKETS) {
      m_per_node_data_pkt_count[ap->u.data_p.seq_no] ++;
      //if (is_src == false)
      //  printf ("[link-data-rcv] [ %d %d ] at %f seq %d\n", ap->src, id, Scheduler::Clock(), ap->u.data_p.seq_no);
    }
    break;

  case JOIN_QUERY:
  case JOIN_FORWARD:
  case JOIN_RESPONSE:
  case CLUSTER_REFRESH:
  case CLUSTER_MERGE:
  case PING_QUERY:
  case PING_RESPONSE:
//slee
	case QUERY_RANDOM:
	case RESPONSE_RANDOM:
	case RETRANSMIT_REQUEST:
	case REQUEST_ADOPT:
//slee

   m_control_byte_count += get_app_packet_size(p);
   m_control_packet_count ++;
//slee
	control_pkt_count++;
//slee
   break;

  default:
    printf ("[Err] Illegal App packet type\n");
  }
  return;
}

void NiceNode::display_on_stop_simulation (void) {

//slee
	if ( id == 1 ) {
		fprintf (stderr, "[StopSim Node %d ] data-pkts %d : control-pkts %d\n", 
			id, data_pkt_count, control_pkt_count);
	}
//slee

/*
  printf ("[StopSim Node %d ] data-pkts %d : control-pkts %d bytes %d\n", id, m_data_packet_count, m_control_packet_count, m_control_byte_count);
//SLEE
  for (int i = 0; i < MAX_DATA_PACKETS; i++)
    if (m_per_node_data_pkt_count[i] > 0)
      printf ("[StopSimData Node %d ] seq-no %d : count %d\n", id, i, m_per_node_data_pkt_count[i]);

*/
  return;
}

#define	BIT_MAP_SIZE_IN_BYTE	8

int NiceNode::get_app_packet_size (Packet *p) {

  AppPacket *ap = (AppPacket *)p;

  switch (ap->st) {

  case JOIN_QUERY:
  case PING_QUERY:
    return 5;

  case JOIN_FORWARD:
    return (2 * HOST_INFO_SIZE +  9);

  case JOIN_RESPONSE:
    return (HOST_INFO_SIZE + 7 + HOST_INFO_SIZE * ap->u.joinresp_p.mbr_count);

  case CLUSTER_REFRESH:
    return (13 + HOST_INFO_SIZE * (ap->u.clusterrefresh_p.mbr_count+ap->u.clusterrefresh_p.hl_mbr_count));

  case CLUSTER_MERGE:
    return (5 + HOST_INFO_SIZE * ap->u.clustermerge_p.mbr_count);

  case PING_RESPONSE:
    return (9 + HOST_INFO_SIZE * ap->u.pingresp_p.mbr_count);

  case PACKET_DATA:
	//extra: seq number & bitmap
    return (2 + HOST_INFO_SIZE + 4 + BITMAP_SIZE/8);
    //return (2 + HOST_INFO_SIZE);

//slee
	case QUERY_RANDOM:
		//original node&agent + up_down/lid
		return HOST_INFO_SIZE + 1;
	case RESPONSE_RANDOM:
		//bitmap + sequence number + answering host
		return BIT_MAP_SIZE_IN_BYTE + 4 + HOST_INFO_SIZE;
	case RETRANSMIT_REQUEST:
		//source + seq + bitmap
		return HOST_INFO_SIZE + BITMAP_SIZE/8 + 4;
		//return HOST_INFO_SIZE + (MAX_BITMAP) * sizeof (int);
	case REQUEST_ADOPT:
		return HOST_INFO_SIZE;
//slee

  default:
    return 0;
  }
  return 0;
}


