/**************************************************
 *
  File: uninode.cc
  Author: Suman Banerjee <suman@cs.umd.edu>
  Date: July 31, 2001
  Terms: GPL

  UNI implementation in myns

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 **************************************************/

#include <stdlib.h>
#include <scheduler.h>
#include <constants.h>
#include "uninode.h"
#include "app-packet.h"

#define MAX_DATA_PACKETS (max_data_packets.getval().u.i)
NamedConstant max_data_packets("max_data_packets",1000);

void * safe_malloc (int size);

UniNode::UniNode (void) : Node () {

  m_data_packet_count = 0;
  m_control_packet_count = 0;
  if (MAX_DATA_PACKETS > 0)
    m_per_node_data_pkt_count = (int *) safe_malloc (sizeof(int) * MAX_DATA_PACKETS)
;
  else
    m_per_node_data_pkt_count = NULL;

  for (int i = 0; i < MAX_DATA_PACKETS; i++)
    m_per_node_data_pkt_count[i] = 0;

}

UniNode::~UniNode (void) {

  if (m_per_node_data_pkt_count != NULL)
    free(m_per_node_data_pkt_count);
}

void UniNode::specific_tx_pkt (Packet *p, int dst) {

  AppPacket *ap = (AppPacket *)p;
  if (ap->st == PACKET_DATA)
    printf ("[link-data-send] [ %d %d ] at %f seq %d\n", id, dst, Scheduler::Clock(), ap->u.data_p.seq_no);
  return;
}

void UniNode::specific_rx_pkt_handler (Packet *p, bool is_src) {

  AppPacket *ap = (AppPacket *)p;

  switch (ap->st) {

  case PACKET_DATA:

    m_data_packet_count ++;
    assert (ap->u.data_p.seq_no >= 0);
    if (ap->u.data_p.seq_no < MAX_DATA_PACKETS) {
      m_per_node_data_pkt_count[ap->u.data_p.seq_no] ++;
      //if (is_src == false)
      //  printf ("[link-data-rcv] [ %d %d ] at %f seq %d\n", ap->src, id, Scheduler::Clock(), ap->u.data_p.seq_no);
    }
    break;

  case JOIN:
  case SOURCE:
  case SOURCE_UPDATE:

   m_control_packet_count ++;
   break;

  default:
    printf ("[Err] Illegal App packet type\n");
  }
  return;
}

void UniNode::display_on_stop_simulation (void) {

  printf ("[StopSim Node %d ] data-pkts %d : control-pkts %d\n", id, m_data_packet_count, m_control_packet_count);
  for (int i = 0; i < MAX_DATA_PACKETS; i++)
    if (m_per_node_data_pkt_count[i] > 0)
      printf ("[StopSimData Node %d ] seq-no %d : count %d\n", id, i, m_per_node_data_pkt_count[i]);

  return;
}

void * safe_malloc (int size) {

  void *p = malloc (size);
  if (p == NULL) {
     perror("malloc:");
     exit(-2);
     return NULL;
  }
  return p;
}

